/*
 *  Plasma.cpp - Simple plasma fire V2.0
 *               Written by Christian Bauer
 *               Public domain
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>
#include <string.h>
#include <stdlib.h>


// Types
typedef unsigned char UBYTE;
typedef unsigned long ULONG;


// Constants
const ulong APP_SIGNATURE = '????';
const int DISPLAY_X = 320;
const int DISPLAY_Y = 240;

// Window/Application messages
const int MSG_REDRAW = 1;


// Global variables
UBYTE *Buffer[2];
sem_id BufferSem[2];

UBYTE PalRed[256];
UBYTE PalGreen[256];
UBYTE PalBlue[256];

UBYTE ColorConv[256];


class Renderer;
class PlasmaWindow;
class BitmapView;


// Application object
class Plasma : public BApplication {
public:
	Plasma();
	virtual void ReadyToRun(void);
	virtual bool QuitRequested(void);
	virtual void MessageReceived(BMessage *msg);

private:
	void recalc(void);

	int draw_buffer;
	PlasmaWindow *the_window;
};


// Window object
class PlasmaWindow : public BWindow {
public:
	PlasmaWindow(int width, int height);
	virtual bool QuitRequested(void);
	virtual void MessageReceived(BMessage *msg);

private:
	BBitmap *the_bitmap;
	BitmapView *main_view;
	int xmod;
};


// Bitmap view
class BitmapView : public BView {
public:
	BitmapView(BRect frame, BBitmap *bitmap);
	virtual void Draw(BRect update);

private:
	BBitmap *the_bitmap;
};


// Create application object and start it
int main(int argc, char **argv)
{	
	Plasma *the_app;

	srand(system_time());

	the_app = new Plasma();
	the_app->Run();
	delete the_app;

	return 0;
}


/*
 *  Application constructor
 */

Plasma::Plasma() : BApplication(APP_SIGNATURE)
{
	Buffer[0] = Buffer[1] = NULL;
	the_window = NULL;
}


/*
 *  Open window and renderer
 */

void Plasma::ReadyToRun(void)
{
	int i;

	// Allocate buffers
	Buffer[0] = new UBYTE[DISPLAY_X * (DISPLAY_Y + 3)];
	Buffer[1] = new UBYTE[DISPLAY_X * (DISPLAY_Y + 3)];
	memset(Buffer[0], 0, DISPLAY_X * (DISPLAY_Y + 3));
	memset(Buffer[1], 0, DISPLAY_X * (DISPLAY_Y + 3));
	draw_buffer = 1;

	// Create buffer semaphores
	BufferSem[0] = create_sem(1, "Buffer 0");
	BufferSem[1] = create_sem(1, "Buffer 1");

	// Create color palette
	for (i=0; i<256; i++)
		PalRed[i] = PalGreen[i] = PalBlue[i] = 0;
	for (i=0; i<8; i++)		// 0..7: increase blue
		PalBlue[i] = i << 3;
	for (i=0; i<8; i++)		// 8..15: decrease blue
		PalBlue[i+8] = (7-i) << 3;
	for (i=8; i<32; i++)	// 8..31: increase red
		PalRed[i] = (i-8) * 255 / (32-8+1);
	for (i=32; i<56; i++) {	// 32..55: increase green
		PalRed[i] = 255;
		PalGreen[i] = (i-32) * 255 / (56-32+1);
	}
	for (i=56; i<80; i++) {	// 56..79: increase blue
		PalRed[i] = PalGreen[i] = 255;
		PalBlue[i] = (i-56) * 255 / (80-56+1);
	}
	for (i=80; i<256; i++)
		PalRed[i] = PalGreen[i] = PalBlue[i] = 255;

	// Create color conversion table
	for (i=0; i<256; i++)
		ColorConv[i] = index_for_color(PalRed[i], PalGreen[i], PalBlue[i]);

	// Open window
	the_window = new PlasmaWindow(DISPLAY_X, DISPLAY_Y);

	// Start action
	PostMessage(MSG_REDRAW);
}


/*
 *  Quit requested (either by menu or by closing the window)
 */

bool Plasma::QuitRequested(void)
{
	// Make sure the window closes first
	if (BApplication::QuitRequested()) {

		// Delete semaphores
		acquire_sem(BufferSem[0]);
		acquire_sem(BufferSem[1]);
		delete_sem(BufferSem[0]);
		delete_sem(BufferSem[1]);

		// Deallocate buffers
		if (Buffer[0])
			delete[] Buffer[0];
		if (Buffer[1])
			delete[] Buffer[1];

		return TRUE;
	}
	return FALSE;
}


/*
 *  Handles redraw messages
 */

void Plasma::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_REDRAW:

			// Recalc buffer
			acquire_sem(BufferSem[draw_buffer]);
			recalc();
			release_sem(BufferSem[draw_buffer]);

			// Tell window to convert and redraw the new buffer
			BMessage *winmsg = new BMessage(MSG_REDRAW);
			winmsg->AddLong("buffer", draw_buffer);
			the_window->PostMessage(winmsg);

			// Switch to other buffer
			draw_buffer ^= 1;

			// Start over
			PostMessage(MSG_REDRAW);
			break;

		default:
			inherited::MessageReceived(msg);
	}
}


/*
 *  Main calculation (write to Buffer[draw_buffer])
 */

void Plasma::recalc(void)
{
	UBYTE *p, *q, *r;
	int x, y;
	int decr;

	// Copy up by 1 pixel and interpolate
	p = Buffer[draw_buffer ^ 1] + 2 * DISPLAY_X; q = Buffer[draw_buffer];
	decr = 1;
	for (y=0; y<DISPLAY_Y; y++) {
		for (x=0; x<DISPLAY_X; x++) {
			ULONG sum;
			sum = p[-DISPLAY_X-1];
			sum += p[-DISPLAY_X];
			sum += p[-DISPLAY_X+1];
			sum += p[-1];
			sum += p[1];
			sum += p[DISPLAY_X-1];
			sum += p[DISPLAY_X];
			sum += p[DISPLAY_X+1];
			sum >>= 3;
			if (sum)
				sum -= decr;	// Decrement every second row
			*q++ = sum;
			p++;
		}
		decr ^= 1;
	}

	// Fill lower 3 rows (invisible) with random pixels
	p = Buffer[draw_buffer] + DISPLAY_X * DISPLAY_Y + (DISPLAY_X-256) / 2;
	q = p + DISPLAY_X;
	r = q + DISPLAY_X;
	for (x=0; x<256; x++) {
		*p++ = (rand() & 0xf) + 0x40;
		*q++ = (rand() & 0xf) + 0x40;
		*r++ = (rand() & 0xf) + 0x40;
	}

	// Insert random bright spots
	q = Buffer[draw_buffer] + DISPLAY_X * (DISPLAY_Y+1) + (DISPLAY_X-256) / 2;
	int num_spots = rand() & 0xf;
	for (x=0; x<num_spots; x++) {
		p = q + (rand() & 0xff);
		p[-DISPLAY_X-1] = p[-DISPLAY_X] = p[-DISPLAY_X+1]
			= p[-1] = p[0] = p[1]
			= p[DISPLAY_X-1] = p[DISPLAY_X] = p[DISPLAY_X+1] = 255;
	}
}


/*
 *  Window constructor
 */

PlasmaWindow::PlasmaWindow(int width, int height) : BWindow(BRect(0, 0, width-1, height-1), "Plasma", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	// Allocate bitmap
	the_bitmap = new BBitmap(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1), B_COLOR_8_BIT);
	xmod = the_bitmap->BytesPerRow();

	// Move window to right position
	MoveTo(80, 60);

	// Create bitmap view
	Lock();
	main_view = new BitmapView(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1), the_bitmap);
	AddChild(main_view);
	Unlock();

	// Show the window
	Show();
}


/*
 *  Closing the window quits the program
 */

bool PlasmaWindow::QuitRequested(void)
{
	// Free bitmap
	main_view->Sync();
	if (the_bitmap)
		delete the_bitmap;

	// Quit program
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
}


/*
 *  Handles redraw messages
 */

void PlasmaWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_REDRAW:

			// Prevent messages from piling up
			BMessage *msg2;
			MessageQueue()->Lock();
			while ((msg2 = MessageQueue()->FindMessage(MSG_REDRAW, 0)) != NULL)
				MessageQueue()->RemoveMessage(msg2);
			MessageQueue()->Unlock();

			// Find the number of the buffer to redraw
			int draw_buffer = msg->FindLong("buffer");

			// Convert buffer to bitmap
			acquire_sem(BufferSem[draw_buffer]);
			UBYTE *p = Buffer[draw_buffer];
			UBYTE *q = (UBYTE *)the_bitmap->Bits();
			for (int y=0; y<DISPLAY_Y; y++) {
				for (int x=0; x<DISPLAY_X; x++)
					q[x] = ColorConv[p[x]];
				p += DISPLAY_X;
				q += xmod;
			}
			release_sem(BufferSem[draw_buffer]);

			// Draw bitmap
			Lock();
			main_view->Draw(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1));
			Unlock();
			break;

		default:
			inherited::MessageReceived(msg);
	}
}


/*
 *  Bitmap view constructor
 */

BitmapView::BitmapView(BRect frame, BBitmap *bitmap) : BView(frame, "", B_FOLLOW_NONE, B_WILL_DRAW)
{
	the_bitmap = bitmap;
}


/*
 *  Blit the bitmap
 */

void BitmapView::Draw(BRect update)
{
	DrawBitmap(the_bitmap, update, update);
}
